##############################################################################
#
# SubPlone - Simple, nested sites to go
# Copyright (C) 2005 Sidnei da Silva and contributors
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
##############################################################################
"""
$Id: proxy.py 8671 2005-05-23 18:39:04Z dreamcatcher $
"""

import sys
from Globals import Persistent
from Acquisition import Implicit, aq_inner, aq_parent, aq_base
from AccessControl.Owned import Owned
from AccessControl import getSecurityManager, Unauthorized
from AccessControl.Role import RoleManager
from BTrees.OOBTree import OOBTree, OOSet
from BTrees.Length import Length
from OFS.ObjectManager import BeforeDeleteException
from OFS.Traversable import Traversable
from ZODB.POSException import ConflictError
from Products.CMFCore.utils import getToolByName
from Products.SubPlone.config import *
from Products.SubPlone.permissions import AddReference
from Products.SubPlone.registry import RefRegistry

class ForcedRef:

    def deleteHook(self, context):
        target = self.resolve(context, marker)
        if target is marker:
            # Target is gone or couldn't be found.
            return
        security = getSecurityManager()
        if not security.checkPermission(AddReference, target):
            raise Unauthorized, 'You are not allowed to remove this reference'

class PathRef(Persistent):

    def __init__(self, path, *args, **kw):
        self.path = path

    def resolve(self, context, default=None):
        # Assume references are from the portal's root.
        purl = getToolByName(context, 'portal_url')
        root = purl.getPortalObject()
        target = root.restrictedTraverse(self.path, default)
        if target is default:
            return default
        return target.__of__(context)

    def deleteHook(self, context):
        pass

RefRegistry.register('PathRef', 'Path Reference', PathRef)

class ForcedPathRef(ForcedRef, PathRef):
    """Forced Path Reference

    User must have the permission to add references on the target to
    be able to remove a existing reference. If the target is gone
    though, it will be possible to remove the reference.
    """

RefRegistry.register('ForcedPathRef', 'Forced Path Reference', ForcedPathRef)

class UIDRef(Persistent):

    def __init__(self, uid, *args, **kw):
        assert uid is not None, 'UID must not be None'
        self.uid = uid

    def resolve(self, context, default=None):
        # Assume references are from the portal's root.
        rc = getToolByName(context, 'reference_catalog')
        obj = rc.lookupObject(self.uid)
        if obj is not None:
            return obj.__of__(context)
        return default

    def deleteHook(self, context):
        pass

RefRegistry.register('UIDRef', 'UID Reference', UIDRef)

class ForcedUIDRef(ForcedRef, UIDRef):
    """Forced UID Reference

    User must have the permission to add references on the target to
    be able to remove a existing reference. If the target is gone
    though, it will be possible to remove the reference.
    """

RefRegistry.register('ForcedUIDRef', 'Forced UID Reference', ForcedUIDRef)

def parent(obj):
    return aq_parent(aq_inner(obj))

class ProxyStorage(Persistent, Implicit):

    def __init__(self):
        self.by_id = OOBTree()
        self.by_meta = OOBTree()
        self.metadata = OOBTree()
        self.count = Length()

    def add(self, key, ref):
        # Remove existing refs before trying to add, that should
        # prevent the user from overriding 'Forced' refs for example.
        if self.has_key(key):
            self.remove(key)

        # Resolve the reference early to get the target's meta_type,
        # which is then used to build an internal structure to speedup
        # filtering on objectIds and friends.
        self.by_id[key] = ref
        context = parent(self)
        target = ref.resolve(context)
        meta = getattr(aq_base(target), 'meta_type')

        if not self.by_meta.has_key(meta):
            self.by_meta[meta] = OOSet()
        self.by_meta[meta].insert(key)
        self.metadata[key] = {'meta_type':meta}
        self.count.change(1)

    def remove(self, key):
        # Call the deleteHook. If the user is not allowed to delete
        # this reference, then a Unauthorized exception will be
        # raised.
        ref = self.by_id[key]
        ref.deleteHook(self)
        md = self.metadata[key]
        meta = md['meta_type']
        del self.metadata[key]
        self.by_meta[meta].remove(key)
        del self.by_id[key]
        self.count.change(-1)

    def has_key(self, key):
        return self.by_id.has_key(key)

    def items(self):
        return self.by_id.items()

    def keys(self):
        return self.by_id.keys()

    def meta(self, _meta):
        if not self.by_meta.has_key(_meta):
            return ()
        return self.by_meta[_meta]

    def get(self, key, default=None):
        return self.by_id.get(key, default)

class ProxyContainer(Persistent, Traversable, Implicit, RoleManager):

    def __init__(self):
        self.proxy = ProxyStorage()

    def objectIds(self, spec=None):
        # Returns a list of subobject ids of the current object.
        # If 'spec' is specified, returns objects whose meta_type
        # matches 'spec'.
        if spec is not None:
            if isinstance(spec, basestring):
                spec = [spec]
            set = []
            for meta in spec:
                set.extend(self.proxy.meta(meta))
            return set
        return list(self.proxy.keys())

    def _getOb(self, id, default=marker):
        ref = self.proxy.get(id, marker)
        if ref is marker:
            return default
        return ref.resolve(self, default)

    def _delOb(self, id):
        if not self.proxy.has_key(id):
            return marker
        self.proxy.remove(id)

    def _delObject(self, id, dp=1):
        # dp is here for backwards compatibility as it once
        # allowed to override the use of the delete protocol.
        # See http://collector.zope.org/Zope/1377
        if not self.proxy.has_key(id):
            return marker
        try:
            self.proxy.remove(id)
        except BeforeDeleteException, ob:
            raise
        except Unauthorized:
            raise
        except ConflictError:
            raise
        except:
            log('ProxyStorage.remove() threw',
                error=sys.exc_info())
            pass
        self._objects = tuple(filter(lambda i, n=id: i['id'] !=n , self._objects))

    def objectValues(self, spec=None):
        # Returns a list of actual subobjects of the current object.
        # If 'spec' is specified, returns only objects whose meta_type
        # match 'spec'.
        return [ProxyContainer._getOb(self, id)
                for id in ProxyContainer.objectIds(self, spec)]

    def objectItems(self, spec=None):
        # Returns a list of (id, subobject) tuples of the current object.
        # If 'spec' is specified, returns only objects whose meta_type match
        # 'spec'
        return [(id, ProxyContainer._getOb(self, id))
                for id in ProxyContainer.objectIds(self, spec)]

    def __getitem__(self, key):
        return self._getOb(key)

    # The two methods below, for playing nicely with BTreeFolder

    def objectCount(self):
        return self.proxy.count()

    def has_key(self, key):
        return self.proxy.has_key(key)
