##############################################################################
#
# SubPlone - Simple, nested sites to go
# Copyright (C) 2005 Sidnei da Silva and contributors
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
##############################################################################
"""
$Id: utils.py 9036 2005-06-10 21:51:17Z dreamcatcher $
"""

from config import log, INFO
from inspect import getargs
import ExtensionClass
from ComputedAttribute import ComputedAttribute
from Acquisition import ImplicitAcquisitionWrapper as IAW
from Products.Archetypes.utils import mapply
from Products.SubPlone.config import marker

class DefaultView:

    default_view = None

    def __browser_default__(self, request):
        default = getattr(self, 'default_view', None) or 'view'
        return self, (default,)

class ContentPermMap(dict):

    def __setitem__(self, key, value):
        if key in self:
            values = self[key]
            if isinstance(values, str):
                values = [values]
            elif isinstance(values, tuple):
                values = list(value)

            if value is None or values is None:
                values = None
            elif isinstance(value, str ):
                values.append( value )
            elif isinstance(value, (list, tuple)):
                values.extend( list(value ) )
            else:
                raise SyntaxError("Unknown %r %r%"(values, value))
            value = values

        super(ContentPermMap, self).__setitem__(key, value)


def groupTypesByPerm(at_types,
                     content_types,
                     constructors,
                     permission_map):

    res = {}

    default_perm = None
    used_types = []


    for permission in permission_map:
        portal_types = permission_map[permission]

        if portal_types is None:
            # default perm
            default_perm = permission
            continue

        types_for_perm = res.setdefault(permission, [])

        for idx, atype in enumerate(at_types):
            pt = atype['klass'].portal_type
            if pt in portal_types:
                assert pt not in used_types
                types_for_perm.append(
                    (content_types[idx],
                     constructors[idx])
                    )
                used_types.append(pt)


    if not default_perm:
        return res

    # handle default add permission
    types_for_perm = res.setdefault(default_perm, [])

    for idx, atype in enumerate(at_types):
        pt = atype['klass'].portal_type
        if not pt in used_types:
            types_for_perm.append(
                (content_types[idx],
                 constructors[idx])
                )

    return res

def field_position(schema, fname):
    return schema._names.index(fname)

def reorder_field(schema, fname, pos):
    names = schema._names
    names.remove(fname)
    names.insert(pos, fname)

def decorate(field, get_dict=lambda f: f.__dict__, **kwargs):
    """ Given a Field instance, returns back a newly decorated instance
    This potentially could cause the Sun to implode - be mindful.
    """

    decorated = field.copy()
    d = get_dict(decorated)
    d.update(kwargs)
    return decorated

def decorateField(field, **kwargs):
    return decorate(field, **kwargs)

def decorateWidget(field, **kwargs):
    return decorate(field, get_dict=lambda f: f.widget.__dict__, **kwargs)

def decorateSchema(schema, fnames, func=decorateField, **kwargs):
    for name in fnames:
        schema[name] = func(schema[name], **kwargs)

def decorateSchemaFields(schema, fnames, func=decorateField, **kwargs):
    decorateSchema(schema, fnames, func=func, **kwargs)

def decorateSchemaWidgets(schema, fnames, func=decorateWidget, **kwargs):
    decorateSchema(schema, fnames, func=func, **kwargs)

class StopProcessing(Exception): pass

class Pipe_(ExtensionClass.Base):
    """A method pipeline.

    Basically, what it does is to act as a wrapper around an existing
    method to allow plugging in any number of pre and post functions
    to be called and also an aggregator to act upon the results.

    >>> class agg:
    ...    def __init__(self):
    ...        self.result = []
    ...    def __call__(self, name, res):
    ...        self.result.append((name, res))
    ...    def start(self):
    ...        self.__init__()
    ...    end = start

    >>> class base(ExtensionClass.Base):
    ...    def foo(self, bar=None, baz=None):
    ...        return bar, baz

    >>> class piped(base):
    ...    def pre1(self, bar=0, baz=1):
    ...        return bar, baz
    ...    def pre2(self, bar=1, baz=2):
    ...        return bar, baz
    ...    def pos1(self, bar=2, baz=1):
    ...        return bar, baz
    ...    foo = Pipe_('foo', base, agg(),
    ...               before=[('pre1', pre1),
    ...                       ('pre2', pre2)],
    ...               after=[('pos1', pos1)])

    >>> from pprint import pprint
    >>> pprint(piped().foo(), width=60)
    [('pre1', (0, 1)),
     ('pre2', (1, 2)),
     ('foo', (None, None)),
     ('pos1', (2, 1))]

    >>> pprint(piped().foo(bar=4), width=60)
    [('pre1', (4, 1)),
     ('pre2', (4, 2)),
     ('foo', (4, None)),
     ('pos1', (4, 1))]

    >>> pprint(piped().foo(a=1, baz=4), width=60)
    [('pre1', (0, 4)),
     ('pre2', (1, 4)),
     ('foo', (None, 4)),
     ('pos1', (2, 4))]

    >>> def pos2(self, bar=1, baz=0):
    ...     return bar, baz

    >>> piped.foo.after.append(('pos2', pos2))

    >>> pprint(piped().foo(), width=60)
    [('pre1', (0, 1)),
     ('pre2', (1, 2)),
     ('foo', (None, None)),
     ('pos1', (2, 1)),
     ('pos2', (1, 0))]

    >>> pprint(piped().foo(bar=4), width=60)
    [('pre1', (4, 1)),
     ('pre2', (4, 2)),
     ('foo', (4, None)),
     ('pos1', (4, 1)),
     ('pos2', (4, 0))]

    >>> def stop(self):
    ...     raise StopProcessing

    >>> piped.foo.before.insert(1, ('stop', stop))
    >>> pprint(piped().foo(), width=60)
    [('pre1', (0, 1))]
    """

    def __init__(self, name, target, aggregator,
                 before=None, after=None):
        self.name = name
        self.target = target
        self.aggregator = aggregator
        if before is None:
            before = []
        if after is None:
            after = []
        self.before = before
        self.after = after

    def function(self, klass, name):
        return getattr(klass, name).im_func

    def __call__(self, instance, *args, **kw):
        # Call the 'before' functions, then the original one, then the
        # 'after' functions.
        tname = self.target.__name__
        args = (instance,) + args
        log('Calling %s of %s on %r with %r' %
            (self.name, tname, instance, (args, kw)), level=INFO)
        func = self.function(self.target, self.name)
        functions = self.before + [(self.name, func)] + self.after
        agg = self.aggregator
        agg.start()
        for fid, func in functions:
            try:
                res = mapply(func, *args, **kw)
                agg(fid, res)
            except StopProcessing:
                break
        res = agg.result
        agg.end()
        return res

    class wrapper:
        """A wrapper for pipe
        """
        def __init__(self, pipe, instance):
            # Using a list avoids touching acquisition.
            self.pipe = [pipe]
            self.instance = [instance]


        def __call__(self, *args, **kw):
            return self.pipe[0](self.instance[0], *args, **kw)

        def __of__(self, instance):
            return self

    def __of__(self, instance):
        return IAW(self.wrapper(self, instance), instance)


def Pipe(name, target, aggregator,
         before=None, after=None):
    def method(self):
        pipe = Pipe_(name, target, aggregator,
                     before, after)
        return pipe.__of__(self)
    return ComputedAttribute(method, 1)

class ListAggregator:

    def __init__(self):
        self.result = []

    start = end = __init__

    def __call__(self, key, value):
        self.result.extend(value)

class LastAggregator:

    def __init__(self):
        self.result = None

    start = end = __init__

    def __call__(self, key, value):
        self.result = value

class FirstAggregator:

    def __init__(self):
        self.result = marker

    start = end = __init__

    def __call__(self, key, value):
        if value is not marker:
            self.result = value
            raise StopProcessing

class SumAggregator:

    def __init__(self):
        self.result = 0

    start = end = __init__

    def __call__(self, key, value):
        self.result += value

class OrAggregator:

    def __init__(self):
        self.result = False

    start = end = __init__

    def __call__(self, key, value):
        self.result = self.result or value
        if self.result:
            raise StopProcessing

